(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Position imports Main begin

datatype pos = Empty ("\<epsilon>") | PCons "nat" "pos" (infixr "<#" 65)

primrec append :: "pos \<Rightarrow> pos \<Rightarrow> pos" (infixr "@" 65)
where append_Empty: "\<epsilon>@q = q"
    | append_PCons:  "(i<#p)@q = i<#(p@q)"

abbreviation singleton :: "nat \<Rightarrow> pos" ("<_>") where "<i> == i<#\<epsilon>"
abbreviation snoc :: "pos \<Rightarrow> nat \<Rightarrow> pos" (infixl "#>" 65) where "p#>i == p@<i>"

lemma not_PCons_self[simp]: "p \<noteq> i<#p" by (induct p) auto
lemmas not_PCons_self2[simp] = not_PCons_self[symmetric]
lemma not_snoc_self[simp] : "p \<noteq> p#>i" by (induct p) auto
lemmas not_snoc_self2[simp] = not_snoc_self[symmetric]
lemma impossible_PCons: "size p \<le> size q \<Longrightarrow> p = i<#q = False" by (induct p) auto
lemma size_0_conv[iff]: "(0 = size p) = (p = \<epsilon>)" by (cases p) auto
lemma size_greater_0_conf[iff]: "(0 < size p) = (p \<noteq> \<epsilon>)" by (cases p) auto 

lemma append_assoc[simp]: fixes p q r :: "pos" shows "(p@q)@r = p@(q@r)" by (induct p) auto
lemma append_Empty2[simp]: "p@\<epsilon> = p" by (induct p) auto
lemma [iff]: fixes p :: "pos" shows "size(p@q) = size p + size q" by (induct p) auto

interpretation semigroup_append: semigroup_add ["op @ :: pos \<Rightarrow> pos \<Rightarrow> pos"] by unfold_locales simp
interpretation monoid_append: monoid_add ["\<epsilon>" "op @ :: pos \<Rightarrow> pos \<Rightarrow> pos"] by unfold_locales simp+

lemma append_is_Empty_conv[iff]: "(p@q = \<epsilon>) = (p = \<epsilon> \<and> q = \<epsilon>)" by (induct p) auto
lemma Empty_is_append_conv[iff]: "(\<epsilon> = p@q) = (p = \<epsilon> \<and> q = \<epsilon>)" by (induct p) auto
lemma append_self_conv[iff]: "(p@q = p) = (q = \<epsilon>)" by (induct p) auto
lemma self_append_conv[iff]: "(p = p@q) = (q = \<epsilon>)" by (induct p) auto

lemma append_eq_append_conv[simp,noatp]:
 fixes p q r s :: "pos"
 assumes "size p = size q \<or> size r = size s"
 shows "(p@r = q@s) = (p = q \<and> r = s)"
using assms proof (induct p arbitrary: q)
 case Empty thus ?case by (cases q,simp,force)
next
 case (PCons i p) thus ?case by (cases q,force,simp)
qed

lemma same_append_eq[iff]: fixes p :: "pos" shows "(p@q = p@r) = (q = r)" by simp

primrec rev :: "pos \<Rightarrow> pos"
where "rev \<epsilon> = \<epsilon>"
    | "rev(i<#p) = rev p #> i"

lemma rev_append[simp]: fixes p :: "pos" shows "rev(p@q) = rev q @ rev p" by (induct p) auto
lemma rev_rev_ident[simp]: fixes p :: "pos" shows "rev(rev p) = p" by (induct p) auto

lemma rev_induct[case_names Empty snoc]:
 fixes P :: "pos \<Rightarrow> bool" assumes A: "P \<epsilon>" and B: "\<And>i q. P q \<Longrightarrow> P(q#>i)" 
 shows "P p"
using assms by (simplesubst rev_rev_ident[symmetric],induct rule: pos.induct[where pos = "rev p"]) simp_all

lemma [simp]: "i<#p \<noteq> \<epsilon>" by simp
lemma [simp]: "\<epsilon> \<noteq> i<#p" by simp

lemma [simp]: "p#>i \<noteq> \<epsilon>" by simp
lemma [simp]: "\<epsilon> \<noteq> p#>i" by simp

instantiation pos :: order begin
 definition less_eq_pos :: "pos \<Rightarrow> pos \<Rightarrow> bool" where "less_eq_pos p q == (\<exists>r. (p@r) = q)"
 definition less_pos    :: "pos \<Rightarrow> pos \<Rightarrow> bool" where "less_pos p q    == (p \<le> q \<and> p \<noteq> q)"
 
 instance proof
  fix p q :: pos show "(p < q) = (p \<le> q \<and> p \<noteq> q)" unfolding less_eq_pos_def less_pos_def ..
 next
  fix p :: pos show "(p \<le> p)" unfolding less_eq_pos_def by auto
 next
  fix p q r :: pos assume "p \<le> q" and "q \<le> r" thus "p \<le> r" unfolding less_eq_pos_def by auto
 next
  fix p q :: pos assume "p \<le> q" and "q \<le> p"
  from `p \<le> q` obtain r where "p@r = q" unfolding less_eq_pos_def by best
  from `q \<le> p` obtain s where "q@s = p" unfolding less_eq_pos_def by best
  hence "p@(r@s) = p" unfolding `p@r = q`[symmetric] by auto
  hence "r@s = \<epsilon>" by simp 
  hence "r = \<epsilon>" and "s = \<epsilon>" by auto
  from `p@r = q` show "p = q" unfolding `r = \<epsilon>` by simp
qed
end

lemma prefix_smaller[simp]: fixes p :: "pos" assumes "p < q" shows "size p < size q"
using assms unfolding less_pos_def less_eq_pos_def by auto

lemma size_induct:
 fixes p :: "pos"
 assumes "\<And>p. (\<forall>q. size q < size p \<longrightarrow> P q) \<Longrightarrow> P p" shows "P p"
using assms by (rule measure_induct[of "size"]) iprover

lemma prefix_induct:
 fixes p :: "pos"
 assumes "\<And>p. (\<forall>q<p. P q) \<Longrightarrow> P p" shows "P p"
using assms by (rule measure_induct[of "size"]) auto

interpretation pos: semigroup_mult ["op @ :: pos \<Rightarrow> pos \<Rightarrow> pos"] by (unfold_locales) simp

interpretation pos: monoid_mult ["\<epsilon> :: pos" "op @ :: pos \<Rightarrow> pos \<Rightarrow> pos"] by unfold_locales simp_all

primrec pos_power :: "pos \<Rightarrow> nat \<Rightarrow> pos" (infixl "^" 80)
 where "p ^ 0 = \<epsilon>"
     | "p ^ (Suc n) = p @ (p ^ n)"

interpretation pos: recpower ["\<epsilon>" "op @ :: pos \<Rightarrow> pos \<Rightarrow> pos" "op ^ :: pos \<Rightarrow> nat \<Rightarrow> pos"] by (unfold_locales) simp_all

lemma power_append_distr: "p^(m+n) = p^(m::nat) @ p^n" by (induct m) auto

end
