(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RedPair imports Trs DpFramework begin


definition one_step_ctxt_closed :: "('f,'v)trel \<Rightarrow> bool"
where
  "one_step_ctxt_closed r = (\<forall> f bef s t aft. ((s,t) \<in> r \<longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r))"

lemma one_step_ctxt_closed_imp_ctxt_closed: "one_step_ctxt_closed r \<Longrightarrow> ctxt_closed r"
proof (unfold one_step_ctxt_closed_def ctxt_closed_def)
  assume one: "(\<forall> f bef s t aft. ((s,t) \<in> r \<longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> r))"
  show "CC r \<subseteq> r"
  proof
    fix csct
    assume csct: "csct \<in> CC r"
    thus "csct \<in> r"
    proof (cases csct)
      case (Pair cs ct)
      with csct have "(cs, ct) \<in> CC r" by simp
      thus ?thesis
      proof
	fix s t C 
	assume "cs = C\<langle>s\<rangle>" and "ct = C\<langle>t\<rangle>" and st: "(s,t) \<in> r"
	from st have "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> r" 
	proof (induct C arbitrary: s t)
	  case Hole
	  thus ?case by simp
	next
	  case (More f bef C aft)
	  with one show ?case by auto
	qed
	with prems show ?thesis by simp 
      qed
    qed
  qed
qed

lemma R_inclusion : assumes C: "one_step_ctxt_closed Rel" 
  and s: "sub_closed Rel"
  and orient: "R \<subseteq> Rel" 
  shows "rstep R \<subseteq> Rel"
proof -
  from C have D: "ctxt_closed Rel" by (rule one_step_ctxt_closed_imp_ctxt_closed)
  from orient have "CS R \<subseteq> CS Rel" by (rule CS_monotone)
  with s have "CS R \<subseteq> Rel" by (auto simp: sub_closed_def)
  hence "rstep R \<subseteq> CC Rel" by (rule CC_monotone)
  with D show ?thesis by (auto simp: ctxt_closed_def)
qed
      

lemma compat_tr_compat : fixes S assumes "S O NS \<subseteq> S" shows "S O NS^* \<subseteq> S"
proof 
  fix x
  assume seq: "x \<in> S O NS^*"
  show "x \<in> S"
  proof (cases x)
    case (Pair a b)
    with seq have "(a,b) \<in> S O NS^*" by simp
    with `S O NS \<subseteq> S` have "(a,b) \<in> S" by (rule non_strict_into_strict)
    with Pair show ?thesis by simp
  qed
qed



lemma right_comp_S[simp]: (* SW *)
 assumes "(x, y) \<in> (NS^* O S^* O S \<union> NS^*) O S"
 shows "(x, y) \<in> (NS^* O S^* O S)"
proof-
 from assms have "(x, y) \<in> (NS^* O S^* O S  O S) \<union> (NS^* O S)" by auto
 hence  xy:"(x, y) \<in> (NS^* O (S^* O S) O S) \<union> (NS^* O S)" by auto
 have "S^* O S \<subseteq> S^*" by auto
 with xy have xy:"(x, y) \<in> (NS^* O S^* O S) \<union> (NS^* O S)" by auto
 have "NS^* O S \<subseteq> NS^* O S^* O S" by auto
 with xy show "(x, y) \<in> (NS^* O S^* O S)" by auto
qed

locale redpair =    
  fixes  S :: "('f,'v)trel" ("\<succ>")
    and   NS :: "('f,'v)trel" ("\<succeq>")
  assumes SN: "SN \<succ>"
    and compat: "\<succ> O \<succeq> \<subseteq> \<succ>"
    and wmNS: "one_step_ctxt_closed \<succeq>"
    and subNS: "sub_closed \<succeq>"
    and subS: "sub_closed \<succ>"
begin

lemma mNS: "ctxt_closed \<succeq>" 
  using wmNS by (rule one_step_ctxt_closed_imp_ctxt_closed) 

lemma trCompat : "\<succ> O \<succeq>^* \<subseteq> \<succ>" using compat by (rule compat_tr_compat)

lemma NS_comp_S_is_SN : (* SW *)
 assumes "SN (S :: (('f, 'v) term) brel)"
 shows "SN (NS^* O S^* O S)"
proof
 fix x
 show "SN (NS^* O S^* O S) (x::('f, 'v) term)"
 proof (rule ccontr)
  assume "\<not> SN (NS^* O S^* O S) x"
  hence F: "\<exists>(F :: nat \<Rightarrow> ('f, 'v) term). ((F 0) = x) \<and> ((\<forall>i. (F i, F (Suc i)) \<in> (NS^* O S^* O S)))" unfolding SN_elt_def by auto
  from F obtain F where "F 0 = x" and alli:"\<forall> i. (F i,F (Suc i)) \<in> (NS^* O S^* O S)" by auto
  have "\<forall> i. (\<exists> y z. (F i, y)  \<in> S \<and> (y, z)  \<in> S^* \<and> (z, F (Suc i)) \<in> NS^*)"
   proof
   fix i
   from alli have "(F i,F (Suc i)) \<in> (NS^* O S^* O S)" by auto
   thus "\<exists> y z. (F i, y)  \<in> S \<and> (y, z)  \<in> S^* \<and> (z, F (Suc i)) \<in> NS^*" unfolding rel_comp_def using mem_Collect_eq by auto
   qed
  hence "\<exists> f. (\<forall> i. (\<exists> z. (F i, f i)  \<in> S \<and> ((f i, z)  \<in> S^*) \<and>(z, F (Suc i)) \<in> NS^*))" by (rule choice)
  then obtain f where "\<forall> i. (\<exists> z. (F i, f i)  \<in> S \<and> ((f i, z)  \<in> S^*) \<and>(z, F (Suc i)) \<in> NS^*)" ..
  hence "\<exists> g. \<forall> i. (F i, f i)  \<in> S \<and> (f i, g i)  \<in> S^* \<and> (g i, F (Suc i)) \<in> NS^*" by (rule choice)
  then obtain g where "\<forall> i. (F i, f i)  \<in> S \<and> (f i, g i)  \<in> S^* \<and> (g i, F (Suc i)) \<in> NS^*" ..
  hence "\<forall> i. (f i, g i)  \<in> S^* \<and> (g i, F (Suc i)) \<in> NS^* \<and> (F (Suc i), f (Suc i))  \<in> S" by auto
  hence "\<forall> i. (f i, g i)  \<in> S^* \<and> (g i, f (Suc i))  \<in> S" unfolding rel_comp_def using trCompat by auto
  hence all:"\<forall> i. (f i, g i)  \<in> S^* \<and> (g i, f (Suc i))  \<in> S^+" by auto
  have "\<forall> i. (f i, f (Suc i))  \<in> S^+"
  proof
   fix i
   from all have "(f i, g i)  \<in> S^* \<and> (g i, f (Suc i))  \<in> S^+" ..
   thus "(f i, f (Suc i))  \<in> S^+" using transitive_closure_trans by auto
  qed
  hence "\<exists> x. ((f 0) = x) \<and> (\<forall>i. (f i, f (Suc i)) \<in> S^+)" by auto
  then obtain x where "((f 0) = x) \<and> (\<forall>i. (f i, f (Suc i)) \<in> S^+)" by auto
  hence "\<exists> f. ((f 0) = x) \<and> (\<forall>i. (f i, f (Suc i)) \<in> S^+)" by auto
  hence "\<not> SN (S^+) x" by auto
  hence "\<not> SN (S^+)" by auto
  hence wfSconv:"\<not> wf ((S^+)\<inverse>)" using SN_iff_wf by auto
  from assms have "wf (S\<inverse>)" using SN_imp_wf[where?r=S]  by simp
  with wf_converse_trancl wfSconv show False by auto
 qed
qed

lemma split_NS_S_union: (* SW *)
 assumes "(x, y) \<in> (NS \<union> S)^*"
 shows "(x, y) \<in> NS^* O S^* O S \<union> NS^*"
proof-
 from assms have "\<exists> n. (x, y) \<in> (NS \<union> S)^n" using rtrancl_imp_rel_pow'[where ?R="NS \<union> S"] by auto
 then obtain n where "(x, y) \<in> (NS \<union> S)^n" by auto
 thus "(x, y) \<in> NS^* O S^* O S \<union> NS^*"
 proof (induct n arbitrary: x, simp)
  case (Suc m)
  assume "(x, y) \<in> (NS \<union> S)^(Suc m)"
  hence "\<exists> z. (x, z) \<in> (NS \<union> S) \<and> (z, y) \<in> (NS \<union> S)^m" using rel_pow_Suc_D2[where ?R="NS \<union> S"] by auto
  then obtain z where xz:"(x, z) \<in> (NS \<union> S)" and zy:"(z, y) \<in> (NS \<union> S)^m" by auto
  with Suc have zy:"(z, y) \<in> NS^* O S^* O S \<union> NS^*" by auto
  thus "(x, y) \<in> NS^* O S^* O S \<union> NS^*"
  proof (cases "(x, z) \<in> NS")
   case True
    with zy have "(x, y) \<in> (NS^* O S^* O S  O NS) \<union> (NS^* O NS)" by auto
    hence "(x, y) \<in> (NS^* O S^* O (S  O NS)) \<union> (NS^* O NS)" by auto
    hence "(x, y) \<in> (NS^* O S^* O (S O NS^*)) \<union> (NS^* O NS)" by auto
    with trCompat have xy:"(x, y) \<in> (NS^* O S^* O S) \<union> (NS^* O NS)" by auto
    have "NS^* O NS \<subseteq> NS^*" by auto
    with xy show "(x, y) \<in> (NS^* O S^* O S) \<union> NS^*" by auto
   next
   case False
    with xz have xz:"(x, z) \<in> S" by auto
    with zy have "(x, y) \<in> (NS^* O S^* O S \<union> NS^*) O S" by auto
    thus "(x, y) \<in> (NS^* O S^* O S) \<union> NS^*" using right_comp_S by simp
  qed
 qed
qed



lemma trans_split_NS_S_union: (* SW *)
 assumes "(rstep R)^* \<subseteq> (NS \<union> S)^*"
 shows "(rstep R)^* \<subseteq> NS^* O S^* O S \<union> NS^*"
proof(rule subsetI2)
 fix x y
 assume "(x, y)  \<in> (rstep R)^*"
 with assms have  "(x,y) \<in> (NS \<union> S)^*" by auto
 thus "(x,y) \<in> NS^* O S^* O S \<union> NS^*" using split_NS_S_union by auto
qed

  
lemma redPairSound :
  assumes rules: "R \<subseteq> NS"
  and nonStrict: "rP \<inter> P \<subseteq> NS \<union> S"
  and strict: "P - rP \<subseteq> S"
  and chain: "ichain (P,R) s t \<sigma>"
  shows  "\<exists> j. ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)"
using chain
proof auto
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?next = "\<lambda> i. (s (Suc i), t (Suc i))"
  assume allP: "\<forall> i. (s i, t i) \<in> P"
  assume steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep R)^*"
  have piece1: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> S \<or> (?ts i, ?ts (Suc i)) \<in> NS^* \<and> (?next i) \<in> rP"
  proof
    fix i
    from subNS wmNS rules R_inclusion have "rstep R \<subseteq> NS" by auto
    hence "(rstep R)^* \<subseteq> NS^*" by (rule rtrancl_mono)
    with steps have stepsNS: "(?ts i, ?ss (Suc i)) \<in> NS^*" by blast
    show "(?ts i, ?ts (Suc i)) \<in> S \<or> (?ts i, ?ts (Suc i)) \<in> NS^* \<and> (?next i) \<in> rP" 
    proof (cases "?next i \<in> S \<inter> (P \<union> rP)")
      case True
      hence "(?ss (Suc i), ?ts (Suc i)) \<in> S" using subS by (auto simp: sub_closed_def)
      with stepsNS trCompat show ?thesis by auto 
    next
      case False
      with strict nonStrict allP
      have one: "?next i \<in> NS" and two: "?next i \<in> rP" by auto
      from one have "(?ss (Suc i), ?ts (Suc i)) \<in> NS" using subNS by (auto simp: sub_closed_def)
      with stepsNS have "(?ts i, ?ts (Suc i)) \<in> NS^*" by auto
      with two show ?thesis by simp
    qed
  qed
  
  hence infSeq: "\<forall> i. (?ts i, ?ts (Suc i)) \<in> NS^* \<union> S" by auto
  from SN have "SN S (?ts 0)" by auto
  from infSeq trCompat this have "\<exists> j. \<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> NS^* - S" by (rule non_strict_ending)
  
  from this obtain j where "\<forall> i \<ge> j. (?ts i, ?ts (Suc i)) \<in> NS^* - S" ..
  with piece1 have "\<forall> i \<ge> j. ?next i \<in> rP" by blast
  hence "\<forall> i. (s (i+(Suc j)), t (i + (Suc j))) \<in> rP" by auto
  thus "\<exists> j. \<forall> i. (s (i+j), t (i+j)) \<in> rP" ..
qed

lemma redPairProcessorDpp : 
  assumes rules: "R \<subseteq> NS"
  and nonStrict: "rP \<inter> P \<subseteq> NS \<union> S"
  and strict: "P - rP \<subseteq> S"
  and finite: "finite_dpp (rP, R)"
  shows "finite_dpp (P,R)"
proof (unfold finite_dpp_def, rule notI, (erule exE)+)
  fix s t \<sigma>
  assume min: "min_ichain (P,R) s t \<sigma>"
  hence "ichain (P,R) s t \<sigma>" by simp
  with rules nonStrict strict have "\<exists> j. ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by (rule redPairSound)
  from this obtain j where "ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" ..
  with min have "min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by simp
  with finite show False by (unfold finite_dpp_def, blast)
qed
 
lemma redPairProcessorGdpp : 
  assumes rules: "R \<subseteq> NS"
  and nonStrict: "rP \<inter> nodes P \<subseteq> NS \<union> S"
  and strict: "nodes P - rP \<subseteq> S"
  and finite: "finite_gdpp (node_restriction P rP, R)" (is "finite_gdpp (?rP, R)")
  shows "finite_gdpp (P,R)"
proof -
  have "\<And> s t \<sigma>. min_ichain (nodes P, R) s t \<sigma> \<Longrightarrow> \<exists> j. min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)"
  proof - 
    fix s t \<sigma>
    assume min: "min_ichain (nodes P,R) s t \<sigma>"
    hence "ichain (nodes P,R) s t \<sigma>" by simp
    with rules nonStrict strict have "\<exists> j. ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by (rule redPairSound)
    from this obtain j where "ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" ..
    with min have "min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" by simp
    thus "\<exists> j. min_ichain (rP, R) (shift s j) (shift t j) (shift \<sigma> j)" ..
  qed
  from this and finite show ?thesis by (rule dpp_lift_to_gdpp)
qed

lemma monoRedPairSound : (* SW *)
  assumes monotonic: "one_step_ctxt_closed S"
  and rNonStrict: "rR \<inter> R \<subseteq> NS \<union> S"
  and rStrict: "R - rR \<subseteq> S"
  and pNonStrict: "rP \<inter> P \<subseteq> NS \<union> S"
  and pStrict: "P - rP \<subseteq> S"
  and chain: "ichain (P,R) s t \<sigma>"
  shows  "\<exists> j. ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)"
using chain
proof auto
  let ?ss = "\<lambda> i. s i \<cdot> \<sigma> i"
  let ?ts = "\<lambda> i. t i \<cdot> \<sigma> i"
  let ?nextr = "\<lambda> i. (?ts i, ?ss (Suc i))"
  let ?nextp = "\<lambda> i. (s i, t i)"
  assume allP: "\<forall> i. (s i, t i) \<in> P"
  assume steps: "\<forall> i. (?ts i, ?ss (Suc i)) \<in> (rstep R)^*"
 from monotonic wmNS have monoNSS:"one_step_ctxt_closed (NS \<union> S)" unfolding one_step_ctxt_closed_def by auto
 from subS subNS have subNSS:"sub_closed (NS \<union> S)" unfolding sub_closed_def using cs_union_distr[where ?R="NS" and ?S=S] by auto
 let ?SS = "NS^* O S^* O S"
 from rStrict rNonStrict have "R  \<subseteq> NS \<union> S" by auto
 with monoNSS subNSS R_inclusion have "(rstep R) \<subseteq> (NS \<union> S)" by auto
 hence "(rstep R)^* \<subseteq> (NS \<union> S)^*" by (rule rtrancl_mono)
 hence rsteps_in_NSS_or_NS:"(rstep R)^* \<subseteq> ?SS \<union> NS^*" using trans_split_NS_S_union by auto
 have piece1: "\<forall> i. (?ss i, ?ss (Suc i)) \<in> ?SS \<or> (?ss i, ?ss (Suc i)) \<in> NS^* \<and> ((?nextr i) \<in> (rstep rR)^* \<and> (?nextp i) \<in> rP)"
 proof
  fix i
  show "(?ss i, ?ss (Suc i)) \<in> ?SS \<or> (?ss i, ?ss (Suc i)) \<in> NS^* \<and> (?nextr i) \<in> (rstep rR)^* \<and> (?nextp i) \<in> rP"
   proof (cases "?nextp i \<in> S \<inter> (P \<union> rP)")
    (* case 1: strict decrease with DP step *)
    case True
     hence "(?ss i, ?ts i) \<in> S" using subS by (auto simp: sub_closed_def)
     from steps rsteps_in_NSS_or_NS have "(?nextr i) \<in> ?SS \<union> NS^*" by auto
     with `(?ss i, ?ts i) \<in> S` have "(?ss i, ?ss (Suc i)) \<in> (?SS \<union> NS^*) O S" by auto
     hence "(?ss i, ?ss (Suc i)) \<in> ?SS" by simp
     thus ?thesis by auto
    next
    case False
      with pStrict pNonStrict allP
      have "?nextp i \<in> NS" and "?nextp i \<in> rP" by auto
      hence dp_step_in_NS: "(?ss i, ?ts i) \<in> NS" using subNS by (auto simp: sub_closed_def)
      show "(?ss i, ?ss (Suc i)) \<in> ?SS \<or> (?ss i, ?ss (Suc i)) \<in> NS^* \<and> (?nextr i) \<in> (rstep rR)^* \<and> (?nextp i) \<in> rP"
       proof (cases "(?nextr i) \<in> ((rstep rR)^* - ?SS)")
        (* case 2: strict decrease in R steps *)
        case False
         hence or:"(?nextr i) \<notin> ((rstep rR)^*) \<or> (?nextr i) \<in> ?SS" by auto
         have "(?nextr i) \<in> ?SS"
         proof (cases "(?nextr i) \<in> ?SS", simp)
          case False
           with or steps have "(?nextr i) \<in> ((rstep R)^* ) - ((rstep rR)^* )" by auto
           hence "(?nextr i) \<in> (rstep R)^* O ((rstep R) - (rstep rR)) O (rstep R)^*"
            using rtrancl_diff_decomp[where ?A="rstep R"] by auto
           hence "\<exists> u v. (?ts i, u) \<in> (rstep R)^* \<and> (u, v) \<in> ((rstep R) - (rstep rR)) \<and> (v, ?ss (Suc i)) \<in> (rstep R)^*" by auto
           then obtain u v where
            tu:"(?ts i, u) \<in> (rstep R)^*" and uv:"(u, v) \<in> ((rstep R) - (rstep rR))" and vs:"(v, ?ss (Suc i)) \<in> (rstep R)^*" by auto
           with rsteps_in_NSS_or_NS have tu:"(?ts i, u) \<in> ?SS \<union> NS^*" and vs:"(v, ?ss (Suc i)) \<in> ?SS \<union> NS^*" by auto
           from rStrict monotonic subS R_inclusion have "rstep (R - rR) \<subseteq> S" by auto
           with uv have "(u, v) \<in> S" using rstep_diff by auto
           with tu have "(?ts i, v) \<in> (S O ?SS) \<union> (S O NS^*)" by auto
           with trCompat have "(?ts i, v) \<in> (S O  NS^* O S^* O S) \<union> S" by auto
           with trCompat have "(?ts i, v) \<in> (S O S^* O S) \<union> S" by auto
           hence "(?ts i, v) \<in> S^* O S" by auto
           with vs have nextrstep:"(?ts i, ?ss (Suc i)) \<in> (NS^* O S^* O S O S^* O S) \<union> (NS^* O S^* O S)" by auto
           have "S^* O S O S^* O S \<subseteq> (S^* O S^*) O S" using O_assoc[where ?R="S^*" and ?S="S^*" and ?T="S"] by auto
           hence "S^* O S O S^* O S \<subseteq> S^* O S" using rtrancl_idemp_self_comp by auto
           hence "NS^* O S^* O S O S^* O S \<subseteq> NS^* O S^* O S"
            using rel_comp_mono[where ?r'="NS^*" and ?r="NS^*" and ?s'="S^* O S O S^* O S" and ?s="S^* O S"] by auto
           hence "(NS^* O S^* O S O S^* O S) \<union> (NS^* O S^* O S) \<subseteq> (NS^* O S^* O S)" by auto
           from this nextrstep show  "(?ts i, ?ss (Suc i)) \<in> NS^* O S^* O S" by (rule set_mp)
         qed
         with dp_step_in_NS trCompat show ?thesis by auto
        next
        (* case 3: only weak decrease with both DP and R steps *)
        case True
         hence "(?nextr i) \<in> ((rstep rR)^*)" by auto
         from steps rsteps_in_NSS_or_NS True have "(?nextr i) \<in> NS^*" by auto
         from dp_step_in_NS `(?nextr i) \<in> NS^*` have  "(?ss i, ?ss (Suc i)) \<in> NS^*" by auto
         with `?nextp i \<in> rP` `(?nextr i) \<in> ((rstep rR)^*)` show ?thesis by simp
        qed
       qed
      qed
  hence infSeq: "\<forall> i. (?ss i, ?ss (Suc i)) \<in>  NS^* \<union> (NS^* O S^* O S)" by auto
  from trCompat have compat:"(NS^* O S^* O S) O NS^*  \<subseteq> (NS^* O S^* O S)" by auto
  from SN trCompat have "SN (NS^* O S^* O S)" using NS_comp_S_is_SN by auto
  hence "SN (NS^* O S^* O S) (?ss 0)" by auto
  from infSeq compat this have "\<exists> j. \<forall> i \<ge> j. (?ss i, ?ss (Suc i)) \<in> NS^* - (NS^* O S^* O S)" using non_strict_ending by auto
   from this obtain j where "\<forall> i \<ge> j. (?ss i, ?ss (Suc i)) \<in> NS^* - (NS^* O S^* O S)" ..
   with piece1 have "\<forall> i \<ge> j. (?nextr i \<in> ((rstep rR)^*) \<and> (?nextp i) \<in> rP)" by blast
   hence "\<forall> i. ((?ts (i + j), ?ss (Suc i + j)) \<in> ((rstep rR)^*)  \<and> (s (i + j), t (i + j)) \<in> rP)" by auto
  thus "\<exists>j. (\<forall>i. (s (i + j), t (i + j)) \<in> rP) \<and> (\<forall>i. (t (i + j) \<cdot> \<sigma> (i + j), s (Suc (i + j)) \<cdot> \<sigma> (Suc (i + j))) \<in> (rstep rR)\<^sup>*)" by auto
qed

(*
lemma finiteReducedDpp :
 assumes "rR \<subseteq> R" and "rP \<subseteq> P" and "finite_dpp (P,R)"
 shows "finite_dpp (rP, rR)"
proof (unfold finite_dpp_def, rule notI, (erule exE)+)
  fix s t \<sigma>
  assume min: "min_ichain (rP,rR) s t \<sigma>"
  hence "ichain (rP,rR) s t \<sigma>" by simp
(*  with monotonic rNonStrict rStrict pNonStrict pStrict have "\<exists> j. ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)"
   by (rule monoRedPairSound)
  then obtain j where rchain:"ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)" by auto*)
  from `rR \<subseteq> R` have "CS rR \<subseteq> CS R" by (rule CS_monotone)
  hence "rstep rR \<subseteq> rstep R" by (rule CC_monotone)
  from min have alli:"(\<forall>i. SN (rstep rR) (s i \<cdot> \<sigma> i)) \<and> (\<forall>i. SN (rstep rR) (t i \<cdot> \<sigma> i))" by simp
  have "(\<forall>i. SN (rstep R) (s i \<cdot> \<sigma> i) \<and> SN (rstep R) (t i \<cdot> \<sigma> i))"
  proof
   fix i
   from alli have "SN (rstep rR) (s i \<cdot> \<sigma> i)" and "SN (rstep rR) (t i \<cdot> \<sigma> i)" by auto
   with  `rstep rR \<subseteq> rstep R` have "SN (rstep rR) (s i \<cdot> \<sigma> i)" and "SN (rstep rR) (t i \<cdot> \<sigma> i)"
    using SN_imp_SN_subset[where ?R'="rstep R" and ?rR'="rstep rR"] by auto
   thus " SN (rstep rR) (s i \<cdot> \<sigma> i) \<and> SN (rstep rR) (t i \<cdot> \<sigma> i)" by auto
  qed
  hence "(\<forall>i. SN (rstep rR) (s i \<cdot> \<sigma> i)) \<and> (\<forall>i. SN (rstep rR) (t i \<cdot> \<sigma> i))" by (auto simp:all_conj_distrib)
  with rchain have "min_ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)" using min_ichain.simps by auto
*)

lemma monoRedPairProcessorDpp :
  assumes monotonic: "one_step_ctxt_closed \<succ>"
  and rNonStrict: "rR \<inter> R \<subseteq> NS \<union> S"
  and rStrict: "R - rR \<subseteq> S"
  and pNonStrict: "rP \<inter> P \<subseteq> NS \<union> S"
  and pStrict: "P - rP \<subseteq> S"
  and finite: "finite_dpp (rP, rR)"
  and "rR \<subseteq> R"
  shows "finite_dpp (P,R)"
proof (unfold finite_dpp_def, rule notI, (erule exE)+)
  fix s t \<sigma>
  assume min: "min_ichain (P,R) s t \<sigma>"
  hence "ichain (P,R) s t \<sigma>" by simp
  with monotonic rNonStrict rStrict pNonStrict pStrict have "\<exists> j. ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)"
   by (rule monoRedPairSound)
  then obtain j where rchain:"ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)" by auto
  from `rR \<subseteq> R` have "CS rR \<subseteq> CS R" by (rule CS_monotone)
  hence "rstep rR \<subseteq> rstep R" by (rule CC_monotone)
  from min have alli:"(\<forall>i. SN (rstep R) (s i \<cdot> \<sigma> i)) \<and> (\<forall>i. SN (rstep R) (t i \<cdot> \<sigma> i))" by simp
  have "(\<forall>i. SN (rstep rR) (s i \<cdot> \<sigma> i) \<and> SN (rstep rR) (t i \<cdot> \<sigma> i))"
  proof
   fix i
   from alli have "SN (rstep R) (s i \<cdot> \<sigma> i)" and "SN (rstep R) (t i \<cdot> \<sigma> i)" by auto
   with  `rstep rR \<subseteq> rstep R` have "SN (rstep rR) (s i \<cdot> \<sigma> i)" and "SN (rstep rR) (t i \<cdot> \<sigma> i)"
    using SN_imp_SN_subset[where ?R'="rstep R" and ?rR'="rstep rR"] by auto
   thus " SN (rstep rR) (s i \<cdot> \<sigma> i) \<and> SN (rstep rR) (t i \<cdot> \<sigma> i)" by auto
  qed
  hence "(\<forall>i. SN (rstep rR) (s i \<cdot> \<sigma> i)) \<and> (\<forall>i. SN (rstep rR) (t i \<cdot> \<sigma> i))" by (auto simp:all_conj_distrib)
  with rchain have "min_ichain (rP, rR) (shift s j) (shift t j) (shift \<sigma> j)" using min_ichain.simps by auto
  with finite show False by (unfold finite_dpp_def, blast)
qed


lemma rstepSubsetMono:
 assumes "rR \<subseteq> R"
 shows "rstep rR \<subseteq> rstep R"
proof-
 from assms have "CS rR \<subseteq> CS R" using CS_monotone by auto
 thus "CC (CS rR) \<subseteq> CC (CS R)" using CC_monotone by auto
qed


lemma ruleRemoval:
  assumes monotonic: "one_step_ctxt_closed \<succ>"
  and nonStrict: "R \<subseteq> NS \<union> S"
  and strict: "R - rR \<subseteq> S"
  and subset:"rR \<subseteq> R"
  shows "((SN :: ('f, 'v)trel \<Rightarrow> bool)(rstep R)) = (SN (rstep rR))"
proof
 assume "SN (rstep R)"
 show "SN (rstep rR)"
 proof
  fix  x :: "('f, 'v) term"
  from subset have "rstep rR \<subseteq> rstep R" using rstepSubsetMono by auto
  from `SN (rstep R)` have "SN (rstep R) x" by auto
  from this `rstep rR \<subseteq> rstep R` show "SN (rstep rR) x" by (rule SN_imp_SN_subset)
 qed
next
 assume "SN (rstep rR)"
 show "SN (rstep R)"
 proof
  fix x :: "('f, 'v) term"
  show "SN (rstep R) x"
  proof(rule ccontr)
   assume "\<not> SN (rstep R) x"
   hence "\<exists> f. f 0 = x \<and> (\<forall>i. (f i, f (Suc i)) \<in> rstep R)" using SN_elt_def by auto
   then obtain f where "f 0 = x" and alli:"\<forall>i. (f i, f (Suc i)) \<in> rstep R" by auto
   from SN have "SN S (f 0)" by auto
   from monotonic strict subS R_inclusion have "(rstep (R - rR)) \<subseteq> S" by auto
   from monotonic wmNS have monoNSS:"one_step_ctxt_closed (NS \<union> S)" unfolding one_step_ctxt_closed_def by auto
   from subS subNS have subNSS:"sub_closed (NS \<union> S)" unfolding sub_closed_def using cs_union_distr[where ?R="NS" and ?S=S] by auto
   with nonStrict monoNSS subNSS R_inclusion have "(rstep R) \<subseteq> (NS \<union> S)" by auto
   with `\<forall>i. (f i, f (Suc i)) \<in> rstep R` have "\<forall>i. (f i, f (Suc i)) \<in> (NS \<union> S)" by auto
   with compat `SN S (f 0)` have "\<exists> j. \<forall> i \<ge> j.  (f i, f (Suc i)) \<in> (NS - S)" using non_strict_ending by auto
   then obtain j where stepNS:" \<forall> i \<ge> j.  (f i, f (Suc i)) \<in> (NS - S)" by auto
   have alli:"\<forall> i \<ge> j.  (f i, f (Suc i)) \<in> (rstep rR)"
    proof
     fix i
     show "i \<ge> j \<longrightarrow> (f i, f (Suc i)) \<in> rstep rR"
     proof
      assume "i \<ge> j"
      from alli have "(f i, f (Suc i)) \<in> rstep R" by auto thm rstep_diff
      with `i \<ge> j` stepNS `(rstep (R - rR)) \<subseteq> S` have "(f i, f (Suc i)) \<in>  (rstep R - rstep (R - rR))" by auto
      hence "(f i, f (Suc i)) \<in>  rstep (R - (R - rR))" using rstep_diff[where ?R=R and ?R'="R - rR"] by auto
      from `rR \<subseteq> R` have "R - (R - rR) \<subseteq> rR" by auto
      with `(f i, f (Suc i)) \<in>  rstep (R - (R - rR))`  show "(f i, f (Suc i)) \<in> (rstep rR)"  using rstepSubsetMono by auto
     qed
    qed
   let ?g = "\<lambda> k. f (k + j)"
   from alli have "\<forall> k. (?g k, ?g(Suc k)) \<in> (rstep rR)" and "?g 0 = f j" by auto
   hence "\<exists>g. g 0 = (f j) \<and> (\<forall>k. (g k, g (Suc k)) \<in> (rstep rR))" by best
   with `SN (rstep rR)` show False by auto
  qed
 qed
qed

end

(* now the implementation starts *)

(*
  fixes checkS :: "('f :: sshow, 'v :: sshow)rule \<Rightarrow> sstring list monad"
    and checkNS :: "('f :: sshow, 'v :: sshow)rule \<Rightarrow> sstring list monad"
    and name :: "sstring" 
    and representation :: "sstring list"
  assumes checkSSound: "isOK(checkS rule) \<Longrightarrow> rule \<in> S" 
    and   checkNSSound: "isOK(checkNS rule)  \<Longrightarrow> rule \<in> NS"
begin
*)


fun checkRedPairProc :: "
  (('f :: sshow,'v :: sshow)rule \<Rightarrow> sstring list monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> sstring list monad) \<Rightarrow> 
   sstring \<Rightarrow>
   sstring list \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL  \<Rightarrow> sstring list monad"
where "checkRedPairProc checkSS checkNSS  name representation P rP r  = 
( let (Sp :: ('f,'v) trsL ,NSp) = computeSplit (\<lambda> foo. isOK(checkSS foo)) P in
    (
   checkAllM checkNSS r
    \<leftarrow> (\<lambda> s. ''problem when orienting TRS'' # s ) 
 \<guillemotright> checkAllM checkNSS NSp
    \<leftarrow> (\<lambda> s. ''problem when orienting DPs'' # s )
 \<guillemotright> checkSubset NSp rP
    \<leftarrow> (\<lambda> s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> [''internal error'', ''please contact one of the CeTA members'',''http://cl-informatik.uibk.ac.at/software/ceta'',''ceta@informatik.uibk.ac.at''])
    )
  \<leftarrow> (\<lambda> s. (''could not apply the reduction pair processor with the following''
         # (representation @ ([] # s))))
 )"

fun checkMonoRedPairProc :: "
  (('f :: sshow,'v :: sshow)rule \<Rightarrow> sstring list monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> sstring list monad) \<Rightarrow> 
   sstring \<Rightarrow>
   sstring list \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL  \<Rightarrow>  ('f,'v)trsL  \<Rightarrow> sstring list monad"
where "checkMonoRedPairProc checkSS checkNSS  name representation P rP R rR  = 
( let (Sp :: ('f,'v) trsL , NSp) = computeSplit (\<lambda> foo. isOK(checkSS foo)) P in
  let (Sr :: ('f,'v) trsL , NSr) = computeSplit (\<lambda> foo. isOK(checkSS foo)) R in
    (
   checkAllM checkNSS NSr
    \<leftarrow> (\<lambda> s. ''problem when orienting TRS'' # s ) 
 \<guillemotright> checkAllM checkNSS NSp
    \<leftarrow> (\<lambda> s. ''problem when orienting DPs'' # s )
 \<guillemotright> checkSubset rR R
    \<leftarrow> (\<lambda> s. [''problem: additional ''])
 \<guillemotright> checkSubset NSr rR
    \<leftarrow> (\<lambda> s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> [''internal error'', ''please contact one of the CeTA members'',''http://cl-informatik.uibk.ac.at/software/ceta'',''ceta@informatik.uibk.ac.at''])
 \<guillemotright> checkSubset NSp rP
    \<leftarrow> (\<lambda> s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> [''internal error'', ''please contact one of the CeTA members'',''http://cl-informatik.uibk.ac.at/software/ceta'',''ceta@informatik.uibk.ac.at''])
    )
  \<leftarrow> (\<lambda> s. (''could not apply the reduction pair processor with the following''
         # (representation @ ([] # s))))
 )"

fun checkRuleRemovalProc :: "
  (('f :: sshow,'v :: sshow)rule \<Rightarrow> sstring list monad) \<Rightarrow>
  (('f,'v)rule \<Rightarrow> sstring list monad) \<Rightarrow> 
   sstring \<Rightarrow>
   sstring list \<Rightarrow> 
   ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> sstring list monad"
where "checkRuleRemovalProc checkSS checkNSS  name representation R rR  = 
( let (Sp :: ('f,'v) trsL , NSp) = computeSplit (\<lambda> foo. isOK(checkSS foo)) R in
    (
   checkAllM checkNSS R
    \<leftarrow> (\<lambda> s. ''problem when orienting rules'' # s ) 
 \<guillemotright> checkSubset NSp rR
    \<leftarrow> (\<lambda> s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> [''internal error'', ''please contact one of the CeTA members'',''http://cl-informatik.uibk.ac.at/software/ceta'',''ceta@informatik.uibk.ac.at''])
 \<guillemotright> checkSubset rR R
    \<leftarrow> (\<lambda> s. case checkSS s of Left err \<Rightarrow> err | Right ok \<Rightarrow> [''internal error'', ''please contact one of the CeTA members'',''http://cl-informatik.uibk.ac.at/software/ceta'',''ceta@informatik.uibk.ac.at''])
    )
  \<leftarrow> (\<lambda> s. (''could not apply the rule removal processor with the following''
         # (representation @ ([] # s))))
 )"
end